var fs = require('fs');
var path = require('path');
var debug = require('debug')('chromecast');
var chromecast = require('chromecast-controller');
var internalIp = require('internal-ip').v4;
var bridge = require('./bridge');
var extract = require('./extract');
var gnome = require('./gnome');
var notify = require('./notify');
var controller = require('./remote-controller');
var messages = require('./messages');
var shared = require('../shared');

var playerStatus = {};
var playerVolume = 1;
var remoteBusy = false;
var castInterval;
var playTimeout;
var initType;

exports.cast = function()
{
	remoteBusy = true;

	clearPlayTimeout();
	stopCastInterval();

	if(chromecast._player)
	{
		chromecast._player.removeListener('close', finishCast);
		chromecast._player.removeListener('status', handleChromecastStatus);
	}

	debug('NEW SELECTION');

	if(!extract.subsProcess && !extract.coverProcess)
		return initChromecast();

	debug('Waiting for extract processes to finish...');

	var checkInterval = setInterval(() =>
	{
		/* Cast after extract processes are done */
		if(!extract.subsProcess && !extract.coverProcess)
		{
			debug('Processes finished extracting');

			clearInterval(checkInterval);
			initChromecast();
		}
	}, 100);
}

exports.remote = function(action, value)
{
	if(remoteBusy) return;

	if(!isNaN(value) || typeof value == 'boolean')
		debug(`Signal from remote. ACTION: ${action}, VALUE: ${value}`);
	else
		debug(`Signal from remote. ACTION: ${action}`);

	var position;
	remoteBusy = true;

	var unsetBusy = () => remoteBusy = false;

	switch(action)
	{
		case 'PLAY':
			chromecast.play((err) =>
			{
				if(!err)
				{
					playerStatus.playerState = 'PLAYING';
					bridge.setStatusFile(playerStatus);
				}
				unsetBusy();
			});
			break;
		case 'PAUSE':
			chromecast.pause((err) =>
			{
				if(!err)
				{
					playerStatus.playerState = 'PAUSED';
					bridge.setStatusFile(playerStatus);
				}
				unsetBusy();
			});
			break;
		case 'SEEK':
			position = playerStatus.media.duration * value;
			chromecast.seek(position, (err) =>
			{
				if(!err)
				{
					playerStatus.currentTime = position;
					bridge.setStatusFile(playerStatus);
				}
				unsetBusy();
			});
			break;
		case 'SEEK+':
			position = playerStatus.currentTime + value;
			if(position < playerStatus.media.duration)
			{
				chromecast.seek(position, (err) =>
				{
					if(!err)
					{
						playerStatus.currentTime = position;
						bridge.setStatusFile(playerStatus);
					}
					unsetBusy();
				});
			}
			break;
		case 'SEEK-':
			position = playerStatus.currentTime - value;
			if(position < 0) position = 0;
			chromecast.seek(position, (err) =>
			{
				if(!err)
				{
					playerStatus.currentTime = position;
					bridge.setStatusFile(playerStatus);
				}
				unsetBusy();
			});
			break;
		case 'SKIP+':
		case 'SKIP-':
			playerStatus.currentTime = 0;
			bridge.setStatusFile(playerStatus);
			return closeCast(action);
			break;
		case 'REPEAT':
			controller.repeat = value;
			unsetBusy();
			break;
		case 'STOP':
			controller.repeat = false;
			chromecast.stop((err) =>
			{
				if(err) debug(err);

				closeCast(action);
				unsetBusy();
			});
			break;
		case 'VOLUME':
			chromecast.setVolume(parseFloat(value), (err, volume) =>
			{
				if(!err)
				{
					playerVolume = volume.level;
					playerStatus.volume = playerVolume;
					bridge.setStatusFile(playerStatus);
				}
				unsetBusy();
			});
			break;
		default:
			unsetBusy();
			break;
	}
}

function makeID()
{
	var text = "";
	var possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

	for(var i = 0; i < 10; i++)
	{
		text += possible.charAt(Math.floor(Math.random() * possible.length));
	}

	debug(`Generated new session id: ${text}`);

	return text;
}

function initChromecast()
{
	var mimeType = 'video/*';
	var trackIds = null;
	var mediaTracks = null;
	var ip = internalIp.sync();
	var port = bridge.config.listeningPort;
	var sessionID = makeID();

	initType = 'BUFFERED';

	switch(bridge.selection.streamType)
	{
		case 'VIDEO':
			break;
		case 'MUSIC':
			if(bridge.config.musicVisualizer) initType = 'LIVE';
			else mimeType = 'audio/*';
			break;
		case 'PICTURE':
			mimeType = 'image/*';
			break;
		default:
			initType = 'LIVE';
			break;
	}

	var getTextTrackStyle = () =>
	{
		const subsConf = gnome.getJSON('chromecast-subtitles');
		return { ...shared.chromecast.subsStyle, ...subsConf };
	}

	var getTitle = () =>
	{
		if(mimeType === 'audio/*' && extract.metadata) return extract.metadata.title;
		else if(bridge.selection.title) return bridge.selection.title;
		else return path.parse(bridge.selection.filePath).name;
	}

	switch(mimeType)
	{
		case 'video/*':
			trackIds = [1];
			mediaTracks = {
				textTrackStyle: getTextTrackStyle(),
				tracks: shared.chromecast.tracks,
				metadata: {
					metadataType: 1,
					images: [{url: ''}]
				}
			};
			mediaTracks.tracks[0].trackContentId = `http://${ip}:${port}/subswebplayer?session=${sessionID}`;
			break;
		case 'audio/*':
			trackIds = [];
			mediaTracks = {
				metadata: {
					metadataType: 3,
					images: [{url: `http://${ip}:${port}/cover?session=${sessionID}`}]
				}
			};
			break;
		case 'image/*':
			trackIds = [];
			mediaTracks = {
				metadata: {
					metadataType: 4,
					images: [{url: ''}]
				}
			};
			break;
	}

	mediaTracks.metadata.title = getTitle();
	debug(`Media title: ${mediaTracks.metadata.title}`);

	var getAutoplayState = () =>
	{
		return (bridge.selection.streamType === 'MUSIC'
			&& !bridge.config.musicVisualizer) ? true : false;
	}

	var getChromecastName = () =>
	{
		return (bridge.config.chromecastName) ? bridge.config.chromecastName : null;
	}

	var getChromecastIp = () =>
	{
		if(bridge.config.chromecastName)
		{
			const devices = gnome.getJSON('chromecast-devices');

			if(Array.isArray(devices))
			{
				for(var i = 0; i < devices.length; i++)
				{
					if(
						devices[i].ip
						&& devices[i].name === bridge.config.chromecastName
					) {
						return devices[i].ip;
					}
				}
			}
		}

		return null;
	}

	var media = {
		contentId: `http://${ip}:${port}/cast?session=${sessionID}`,
		contentType: mimeType,
		streamType: initType,
		...mediaTracks
	};

	debug(`Setting media: ${JSON.stringify(media)}`);

	var castOpts = {
		autoplay: getAutoplayState(),
		activeTrackIds: trackIds,
		name: getChromecastName(),
		ip: getChromecastIp()
	};

	debug(`Setting opts: ${JSON.stringify(castOpts)}`);
	launchCast(media, castOpts);
}

function launchCast(media, castOpts)
{
	debug('Casting...');

	chromecast.cast(media, castOpts, (err) =>
	{
		if(err)
		{
			debug(`Could not cast: ${err.message}`);
			showTranslatedError(err, castOpts);
		}
		else
		{
			chromecast._player.once('close', finishCast);
			chromecast._player.on('status', handleChromecastStatus);

			debug('Cast started');

			startPlayback(media.contentType);
		}
	});
}

function startCastInterval()
{
	if(!castInterval)
	{
		castInterval = setInterval(() => getChromecastStatus(), 1000);
		debug('Started status interval');
	}
}

function stopCastInterval()
{
	if(castInterval)
	{
		clearInterval(castInterval);
		castInterval = null;

		debug('Stopped status interval');
	}
}

function clearPlayTimeout()
{
	if(playTimeout)
	{
		clearTimeout(playTimeout);
		playTimeout = null;

		debug('Stopped delayed playback');
	}
}

function startPlayback(mimeType)
{
	remoteBusy = false;

	/* Get startup volume level when not casting picture */
	if(mimeType !== 'image/*')
	{
		chromecast.getVolume((err, volume) =>
		{
			if(!err)
			{
				playerVolume = volume;
				debug(`Obtained volume value: ${volume}`);
			}
			else
			{
				playerVolume = 1;
				debug(`Could not obtain volume value. Current setting: ${playerVolume}`);
			}
		});
	}

	/* Delay playback to allow media buffer a little */
	if(mimeType === 'video/*')
	{
		var play = () =>
		{
			playTimeout = null;

			chromecast.play(err =>
			{
				if(!err)
				{
					debug('Playback started');
					startCastInterval();
					/* Show on refresh is handled in bridge.js */
					if(!gnome.isRemote()) gnome.showRemote(true);
				}
				else
				{
					debug('Could not play!');
					return closeCast();
				}
			});
		}

		debug('Starting delayed playback...');

		/* mimeType video + streamType music = music with visualizer */
		/* Visualizations are 60fps, so Chromecast needs to buffer more to not stutter */
		var delay = (bridge.selection.streamType === 'MUSIC') ?
			shared.chromecast.visualizerBuffer : shared.chromecast.videoBuffer;

			playTimeout = setTimeout(() => play(), delay);
	}
	else
	{
		if(mimeType === 'image/*')
			debug('Showing image');
		else
			debug('Playback autostart');

		startCastInterval();

		/* Show on refresh is handled in bridge.js */
		if(!gnome.isRemote()) gnome.showRemote(true);
	}
}

function getChromecastStatus()
{
	chromecast.getStatus((err, status) =>
	{
		if(err)
		{
			debug(`Chromecast status error: ${err}`);
			return showTranslatedError(err);
		}
		else if(status && typeof status === 'object')
		{
			handleChromecastStatus(status);
		}
	});
}

function handleChromecastStatus(status)
{
	playerStatus = { ...playerStatus, ...status };
	playerStatus.volume = playerVolume;

	if(!playerStatus.media)
	{
		playerStatus.currentTime = 0;
		playerStatus.media = { duration: 1 };
	}

	if(status.playerState === 'IDLE')
	{
		switch(status.idleReason)
		{
			case 'ERROR':
				/* Show error and close */
				showIdleError();
			case 'FINISHED':
				return closeCast();
			default:
				break;
		}
	}

	if(!remoteBusy)
		bridge.setStatusFile(playerStatus);
}

function showIdleError()
{
	debug('Chromecast is IDLE due to ERROR!');

	var info = (initType === 'LIVE') ? messages.chromecast.tryAgain : null;
	notify('Chromecast', messages.chromecast.playError, bridge.selection.filePath, info);
}

function showTranslatedError(err, opts)
{
	var msg = err.message.toLowerCase();
	var info = null;
	debug(err);

	opts = opts || {};

	switch(msg)
	{
		case 'device not found':
			notify('Chromecast', messages.chromecast.notFound);
			break;
		case 'load failed':
			notify('Chromecast', messages.chromecast.loadFailed);
			break;
		case 'connection timeout!':
			info = (opts.ip) ? messages.chromecast.verifyIp : null;
			notify('Chromecast', messages.chromecast.connectFailed, null, info);
			break;
		default:
			debug(`Unhandled message: ${msg}`);
			break;
	}
}

/* Normal close from app */
function closeCast(action)
{
	clearPlayTimeout();
	stopCastInterval();

	chromecast._player.removeListener('close', finishCast);
	chromecast._player.removeListener('status', handleChromecastStatus);

	var currentTrackID = bridge.list.indexOf(bridge.selection.filePath) + 1;
	var listLastID = bridge.list.length;

	if(action)
	{
		if(action == 'SKIP+') return controller.changeTrack(currentTrackID + 1);
		else if(action == 'SKIP-') return controller.changeTrack(currentTrackID - 1);
	}

	if(controller.repeat && currentTrackID === listLastID) return controller.changeTrack(1);
	else if(action !== 'STOP' && currentTrackID < listLastID) return controller.changeTrack(currentTrackID + 1);

	debug('Closing cast session...');
	chromecast.close(err =>
	{
		if(!err) debug('Session closed');
		else debug('Could not close session!');

		gnome.showRemote(false);
		debug('Cast finished!');
	});
}

/* Close by external event */
function finishCast()
{
	clearPlayTimeout();
	stopCastInterval();

	/* 'close' listener is auto removed as it is a 'once' event performed here */
	chromecast._player.removeListener('status', handleChromecastStatus);

	gnome.showRemote(false);
	debug('Cast finished due to close event!');
}
